program CPreProcessor;

{$mode objfpc}{$H+}

uses
  {$IFDEF UNIX}{$IFDEF UseCThreads}
  cthreads,
  {$ENDIF}{$ENDIF}
  Classes, SysUtils, CustApp
  { you can add units after this }
  , FileUtil, calc_utils, PreCompiler, strutils;

type

  { TCPreProcessor }

  TCPreProcessor = class(TCustomApplication)
  protected
    procedure DoRun; override;
  public
    function ParamsLine:ansistring;
    procedure PreCompileFu(ALog: TStringList);
    procedure AfterCompileFu(ALog: TStringList);
    constructor Create(TheOwner: TComponent); override;
    destructor Destroy; override;
    procedure WriteHelp; virtual;
  end;

{ TCPreProcessor }


function Preprocess(ASource: ansistring):AnsiString;
var
  iwords: TSourceWords;
  str: TStringList;
begin
  str := TStringList.Create;
  try
    str.Text:=RemoveComments(ASource);
    Result:=textFromWords(str,GroupWords(SourceWords(str.Text)));
  finally
    str.Free;
  end;
end;

procedure PreProcessDir(ADir: ansistring);
var
  files: TStringList;
  ifile: TStringList;
  i: integer;
  iage: LongInt;
  SFileName: AnsiString;
begin
//  i := 56iage ;
  ifile := TStringList.Create;
  files := FindAllFiles(ADir, '*.pas;*.pp;*.inc;*.lpr', false);
  try
    for i := 0 to files.Count-1 do
    begin
      if not DirectoryExistsUTF8(ADir+'/_ppback') then MkDir(ADir+'/_ppback');
      ifile.LoadFromFile(files[i]);
      iage:=FileAgeUTF8(files[i]);
      if Pos('{precompiled}',ifile.Text) <> 1 then
      begin
        SFileName:=ADir+'/_ppback/'+ExtractFileName(files[i]);
        ifile.SaveToFile(ADir+'/_ppback/'+ExtractFileName(files[i])+'~');
        FileSetDateUTF8(ADir+'/_ppback/'+ExtractFileName(files[i])+'~',iage);
        ifile.Text := '{precompiled}'+Preprocess(ifile.Text);
        ifile.SaveToFile(ADir+'/'+ExtractFileName(files[i]));
        ifile.SaveToFile(ADir+'/_ppback/'+ExtractFileName(files[i])+'~pc');
        FileSetDateUTF8(ADir+'/'+ExtractFileName(files[i]),iage);
      end;
    end;


  finally
    files.Free;
    ifile.Free;
  end;
end;

procedure PostProcessDir(ADir: ansistring);
var
  files: TStringList;
  ifile: TStringList;
  i: integer;
  iage: LongInt;
begin
  ifile := TStringList.Create;
  files := FindAllFiles(ADir, '*.pas;*.pp;*.inc;*.lpr', false);
  try
    for i := 0 to files.Count-1 do
      if FileExists(ADir+'/_ppback/'+ExtractFileName(files[i])+'~') then
      begin
        ifile.LoadFromFile(ADir+'/_ppback/'+ExtractFileName(files[i])+'~');
        iage:=FileAgeUTF8(ADir+'/_ppback/'+ExtractFileName(files[i])+'~');
        if Pos('{precompiled}',ifile.Text) <> 1 then
        begin
          ifile.SaveToFile(ADir+'/'+ExtractFileName(files[i]));
          FileSetDateUTF8(ADir+'/'+ExtractFileName(files[i]),iage);
        end;
      end;


  finally
    files.Free;
    ifile.Free;
  end;
end;


procedure TCPreProcessor.PreCompileFu(ALog: TStringList);
var
  i: integer;
  i2: integer;
  idir: AnsiString;
  iop: array of AnsiString;
  iOperators: TStringList;
begin
  for i := 1 to Paramcount do
    if AnsiPos('-Fu', Params[i]) = 1 then
    begin
      idir:=RightStr(Params[i],Length(Params[i])-3);
      ALog.Add(idir);
      if FileExists(idir+'/operators.ini') then
      begin
        ALog.Add(idir+'/operators.ini');
        iOperators := TStringList.Create;
        try
          iOperators.LoadFromFile(idir+'/operators.ini');
          SetLength(OperatorWords,0);
          for i2 := 0 to iOperators.Count-1 do
          begin
            iop := Split(' ',iOperators[i2]);
            if (Length(iop) = 2) and (Length(iop[0]) > 0) and (Length(iop[1]) > 0) then
              addOperator(iop[0],iop[1]);
            ALog.add('addOperator('''+iop[0]+''','''+iop[1]+''');');
          end;
        finally
          iOperators.free;
        end;
        PreProcessDir(idir);
      end;
    end;

end;

procedure TCPreProcessor.AfterCompileFu(ALog: TStringList);
var
  i: integer;
  idir: AnsiString;
begin
  for i := 1 to Paramcount do
    if AnsiPos('-Fu', Params[i]) = 1 then
    begin
      idir:=RightStr(Params[i],Length(Params[i])-3);
      if FileExists(idir+'/operators.ini') then
        PostProcessDir(idir);
    end;

end;

procedure TCPreProcessor.DoRun;
var
  ErrorMsg: String;
  ilog: TStringList;
begin
//  pointer p := @(exit;);
  // quick check parameters
 { ErrorMsg:=CheckOptions('h','help');
  if ErrorMsg<>'' then begin
    ShowException(Exception.Create(ErrorMsg));
    Terminate;
    Exit;
  end;
  }
  // parse parameters
  if HasOption('h','help') then begin
    WriteHelp;
    Terminate;
    Exit;
  end;
  ilog := TStringList.Create;
  if FileExists('fpc.log') then ilog.LoadFromFile('fpc.log');
  ilog.Add(' ');
  ilog.Add(' -- new execute -- ');
  ilog.Add(ParamsLine);
  PreCompileFu(ilog);
//  PreProcessDir('D:\2\');
  ExecuteProcess(ExtractFilePath(Params[0])+'ppcx64',ParamsLine);
//  PostProcessDir('D:\2\');
  AfterCompileFu(ilog);
  ilog.SaveToFile('fpc.log');
  ilog.free;


  if HasOption('p','pre') then begin
    WriteLn('pre: ',GetOptionValue('path'));
    PreProcessDir(GetOptionValue('path'));
    Terminate;
    Exit;
  end;

  if HasOption('a','after') then begin
    WriteLn('after: ',GetOptionValue('path'));
    PostProcessDir(GetOptionValue('path'));
    Terminate;
    Exit;
  end;



  { add your program here }

  // stop program loop
  Terminate;
end;

function TCPreProcessor.ParamsLine: ansistring;
var
  i:integer;
begin
  //
  Result := '';
  for i := 1 to ParamCount do
    Result += ' '+Params[i];
end;



constructor TCPreProcessor.Create(TheOwner: TComponent);
begin
  inherited Create(TheOwner);
  StopOnException:=True;
end;

destructor TCPreProcessor.Destroy;
begin
  inherited Destroy;
end;

procedure TCPreProcessor.WriteHelp;
begin
  { add your help code here }
  writeln('Usage: ',ExeName,' -h');
end;

var
  Application: TCPreProcessor;
begin
  Application:=TCPreProcessor.Create(nil);
  Application.Run;
  Application.Free;
end.

